"""
Transliteration schema base features.
"""
from typing import Dict, List
from .mapping import LetterMapping, PrevMapping, NextMapping, EndingMapping


class Schema:
    """
    Transliteration schema. Defines the way to translate individual letters.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        name: str,
        mapping: Dict[str, str],
        prev_mapping: Dict[str, str] = None,
        next_mapping: Dict[str, str] = None,
        ending_mapping: Dict[str, str] = None,
        samples: List[List[str]] = None,
        description: str = None,
    ):
        self.name = name
        self.description = description
        self.map = LetterMapping(mapping)
        self.prev_map = PrevMapping(prev_mapping or {})
        self.next_map = NextMapping(next_mapping or {})
        self.ending_map = EndingMapping(ending_mapping or {})
        self.samples = samples or []

    def translate_letter(self, prev: str, curr: str, next_: str) -> str:
        """Translate ``curr`` letter according to schema mappings.

        ``prev`` (the previous letter) and ``next_`` (the next one)
        are taken into consideration according to corresponding mappings.

        """
        letter = self.prev_map.get(prev + curr)
        if letter is None:
            letter = self.next_map.get(curr + next_)
        if letter is None:
            letter = self.map.get(curr, curr)
        return letter

    def translate_ending(self, ending: str) -> str:
        """Translate word ending according to schema mapping."""
        return self.ending_map.get(ending)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    @classmethod
    def load(cls, definition: dict):
        """Load schema from definition."""
        defn = SchemaDefinition(definition)
        defn.parse()
        return Schema(
            name=defn.name,
            description=defn.description,
            mapping=defn.mapping,
            prev_mapping=defn.prev_mapping,
            next_mapping=defn.next_mapping,
            ending_mapping=defn.ending_mapping,
            samples=defn.samples,
        )


# pylint: disable=too-many-instance-attributes
class SchemaDefinition:
    """Translitiration schema definition."""

    def __init__(self, source: dict):
        self.source = source
        self.name = ""
        self.description: None
        self.mapping: Dict[str, str] = {}
        self.prev_mapping = None
        self.next_mapping = None
        self.ending_mapping = None
        self.samples: List[List[str]] = []

    def parse(self):
        """Parse source definition, raising ValueError if necessary."""
        self._parse_attr("name", type_=str, required=True, nonempty=True)
        self._parse_attr("description", type_=str, required=False)
        self._parse_attr("mapping", type_=dict, required=True)
        self._parse_attr("prev_mapping", type_=dict, required=False)
        self._parse_attr("next_mapping", type_=dict, required=False)
        self._parse_attr("ending_mapping", type_=dict, required=False)
        self._parse_samples()

    def _parse_attr(self, name, type_, required, nonempty=False):
        value = self.source.get(name)
        if required and value is None:
            raise ValueError(f"{self.name}: Missing schema {name}")
        if required and nonempty and not value:
            raise ValueError(f"{self.name}: Schema {name} should not be empty")
        if value is not None and not isinstance(value, type_):
            raise ValueError(f"{self.name}: Invalid schema {name}: {value}")
        setattr(self, name, value)

    def _parse_samples(self):
        samples = self.source.get("samples")
        if samples is not None and not isinstance(samples, list):
            raise ValueError(f"{self.name}: Invalid schema samples: {samples}")
        if samples:
            for sample in samples:
                self._raise_on_invalid_sample(sample)
        self.samples = samples

    def _raise_on_invalid_sample(self, sample):
        message = f"{self.name}: Invalid schema sample: {sample}"
        if not isinstance(sample, list):
            raise ValueError(message)
        if len(sample) != 2:
            raise ValueError(message)
        if not isinstance(sample[0], str) or not isinstance(sample[1], str):
            raise ValueError(message)



import os
import timeit
import pytest
from iuliia import engine
from iuliia.schema import Schema


def test_translate():
    schema = Schema("test", {})
    assert engine.translate("Iuliia", schema) == "Iuliia"


def test_mapping():
    schema = Schema("test", {"a": "1", "i": "2", "l": "3", "u": "4"})
    assert engine.translate("Iuliia", schema) == "243221"


def test_prev_mapping():
    schema = Schema("test", {}, prev_mapping={"li": ""})
    assert engine.translate("Iuliia", schema) == "Iulia"


def test_next_mapping():
    schema = Schema("test", {}, next_mapping={"iu": "y"})
    assert engine.translate("Iuliia", schema) == "Yuliia"


def test_ending_mapping():
    schema = Schema("test", {}, ending_mapping={"ia": "ya"})
    assert engine.translate("Iuliia", schema) == "Iuliya"


def test_short_word():
    schema = Schema("test", {})
    assert engine.translate("Iu", schema) == "Iu"


def test_empty_word():
    schema = Schema("test", {})
    assert engine.translate("", schema) == ""


@pytest.mark.parametrize(
    "source,expected",
    [
        ("Hello, mankind!", ["Hello", ", ", "mankind", "!"]),
        ("Привет, человечество!", ["Привет", ", ", "человечество", "!"]),
        (
            "(привет) (привет...) привет? а",
            ["(", "привет", ") (", "привет", "...) ", "привет", "? ", "а"],
        ),
    ],
)
def test_split_sentence(source, expected):
    words = list(engine._split_sentence(source))
    assert words == expected


@pytest.mark.skipif(
    os.getenv("TEST_TIMING") is None, reason="skip timing test until explicitly requested"
)
def test_timing():
    mapping = {chr(i): chr(i) for i in range(ord("a"), ord("z"))}
    schema = Schema("test", mapping)
    source = "The quick brown fox jumps over the lazy dog"
    elapsed_sec = timeit.timeit(lambda: engine.translate(source, schema), number=10000)
    max_sec = 1.0
    assert elapsed_sec < max_sec



"""
Transliterate Cyrillic → Latin in every possible way
"""

# flake8: noqa
from .engine import translate
from .schema import Schema
from .schemas import Schemas

__version__ = "0.11.5"
__all__ = []  # type: ignore

ALA_LC = Schemas.ala_lc.value  # type: ignore
ALA_LC_ALT = Schemas.ala_lc_alt.value  # type: ignore
BGN_PCGN = Schemas.bgn_pcgn.value  # type: ignore
BGN_PCGN_ALT = Schemas.bgn_pcgn_alt.value  # type: ignore
BS_2979 = Schemas.bs_2979.value  # type: ignore
BS_2979_ALT = Schemas.bs_2979_alt.value  # type: ignore
GOST_16876 = Schemas.gost_16876.value  # type: ignore
GOST_16876_ALT = Schemas.gost_16876_alt.value  # type: ignore
GOST_52290 = Schemas.gost_52290.value  # type: ignore
GOST_52535 = Schemas.gost_52535.value  # type: ignore
GOST_7034 = Schemas.gost_7034.value  # type: ignore
GOST_779 = Schemas.gost_779.value  # type: ignore
GOST_779_ALT = Schemas.gost_779_alt.value  # type: ignore
ICAO_DOC_9303 = Schemas.icao_doc_9303.value  # type: ignore
ISO_9_1954 = Schemas.iso_9_1954.value  # type: ignore
ISO_9_1968 = Schemas.iso_9_1968.value  # type: ignore
ISO_9_1968_ALT = Schemas.iso_9_1968_alt.value  # type: ignore
MOSMETRO = Schemas.mosmetro.value  # type: ignore
MVD_310 = Schemas.mvd_310.value  # type: ignore
MVD_310_FR = Schemas.mvd_310_fr.value  # type: ignore
MVD_782 = Schemas.mvd_782.value  # type: ignore
SCIENTIFIC = Schemas.scientific.value  # type: ignore
TELEGRAM = Schemas.telegram.value  # type: ignore
UNGEGN_1987 = Schemas.ungegn_1987.value  # type: ignore
WIKIPEDIA = Schemas.wikipedia.value  # type: ignore
YANDEX_MAPS = Schemas.yandex_maps.value  # type: ignore
YANDEX_MONEY = Schemas.yandex_money.value  # type: ignore



"""
Letter mapping for transliteration schema.
"""


class Mapping:
    """Letter map for transliteration schema."""

    def __init__(self, mapping: dict):
        self.map = mapping.copy()

    def get(self, key, default=None):
        """Return mapped value for ``key`` if key is in the map, else ``default``."""
        return self.map.get(key, default)

    def __str__(self):
        return str(self.map)

    def __repr__(self):
        return repr(self.map)


class LetterMapping(Mapping):
    """Mapping for individual letters."""

    def __init__(self, mapping: dict):
        super().__init__(mapping)
        upper_map = {key.capitalize(): value.capitalize() for key, value in mapping.items()}
        self.map.update(upper_map)


class PrevMapping(Mapping):
    """Mapping for letters with respect to previous sibling."""

    def __init__(self, mapping: dict):
        super().__init__(mapping)
        upper_map_1 = {key.capitalize(): value for key, value in mapping.items()}
        upper_map_2 = {key.upper(): value.capitalize() for key, value in mapping.items()}
        self.map.update(upper_map_1)
        self.map.update(upper_map_2)


class NextMapping(Mapping):
    """Mapping for letters with respect to next sibling."""

    def __init__(self, mapping: dict):
        super().__init__(mapping)
        upper_map_1 = {key.capitalize(): value.capitalize() for key, value in mapping.items()}
        upper_map_2 = {key.upper(): value.capitalize() for key, value in mapping.items()}
        self.map.update(upper_map_1)
        self.map.update(upper_map_2)


class EndingMapping(Mapping):
    """Mapping for word endings."""

    def __init__(self, mapping: dict):
        super().__init__(mapping)
        upper_map = {key.upper(): value.upper() for key, value in mapping.items()}
        self.map.update(upper_map)



import pytest
from iuliia.schema import SchemaDefinition


def test_init():
    source = {"whatever": "42"}
    defn = SchemaDefinition(source)
    assert defn.source == source
    assert defn.name == ""


def test_parse_mapping():
    defn = SchemaDefinition(
        {
            "name": "test",
            "mapping": {"a": "z"},
            "prev_mapping": {"ba": "z"},
            "next_mapping": {"ab": "z"},
            "ending_mapping": {"aa": "zz"},
            "samples": [["aaa", "zzz"], ["baa", "bzz"]],
        }
    )
    defn.parse()
    assert defn.name == "test"
    assert defn.mapping == {"a": "z"}
    assert defn.prev_mapping == {"ba": "z"}
    assert defn.next_mapping == {"ab": "z"}
    assert defn.ending_mapping == {"aa": "zz"}
    assert len(defn.samples) == 2
    assert defn.samples[0] == ["aaa", "zzz"]
    assert defn.samples[1] == ["baa", "bzz"]


def test_invalid_name():
    with pytest.raises(ValueError) as exc:
        defn = SchemaDefinition({})
        defn.parse()
    assert str(exc.value) == ": Missing schema name"


def test_empty_name():
    with pytest.raises(ValueError) as exc:
        defn = SchemaDefinition({"name": ""})
        defn.parse()
    assert str(exc.value) == ": Schema name should not be empty"


def test_invalid_mapping():
    with pytest.raises(ValueError) as exc:
        defn = SchemaDefinition({"name": "test"})
        defn.parse()
    assert str(exc.value) == "test: Missing schema mapping"


def test_empty_mapping():
    defn = SchemaDefinition({"name": "test", "mapping": {}})
    defn.parse()
    assert defn.mapping == {}


def test_invalid_prev_mapping():
    with pytest.raises(ValueError) as exc:
        defn = SchemaDefinition({"name": "test", "mapping": {}, "prev_mapping": "42"})
        defn.parse()
    assert str(exc.value) == "test: Invalid schema prev_mapping: 42"


def test_missing_prev_mapping():
    defn = SchemaDefinition({"name": "test", "mapping": {}})
    defn.parse()
    assert defn.prev_mapping is None


def test_empty_prev_mapping():
    defn = SchemaDefinition({"name": "test", "mapping": {}, "prev_mapping": {}})
    defn.parse()
    assert defn.prev_mapping == {}


def test_invalid_next_mapping():
    with pytest.raises(ValueError) as exc:
        defn = SchemaDefinition({"name": "test", "mapping": {}, "next_mapping": "42"})
        defn.parse()
    assert str(exc.value) == "test: Invalid schema next_mapping: 42"


def test_missing_next_mapping():
    defn = SchemaDefinition({"name": "test", "mapping": {}})
    defn.parse()
    assert defn.next_mapping is None


def test_empty_next_mapping():
    defn = SchemaDefinition({"name": "test", "mapping": {}, "next_mapping": {}})
    defn.parse()
    assert defn.next_mapping == {}


def test_invalid_ending_mapping():
    with pytest.raises(ValueError) as exc:
        defn = SchemaDefinition({"name": "test", "mapping": {}, "ending_mapping": "42"})
        defn.parse()
    assert str(exc.value) == "test: Invalid schema ending_mapping: 42"


def test_missing_ending_mapping():
    defn = SchemaDefinition({"name": "test", "mapping": {}})
    defn.parse()
    assert defn.ending_mapping is None


def test_empty_ending_mapping():
    defn = SchemaDefinition({"name": "test", "mapping": {}, "ending_mapping": {}})
    defn.parse()
    assert defn.ending_mapping == {}


def test_invalid_samples():
    with pytest.raises(ValueError) as exc:
        defn = SchemaDefinition({"name": "test", "mapping": {}, "samples": {}})
        defn.parse()
    assert str(exc.value) == "test: Invalid schema samples: {}"


def test_invalid_sample_not_list():
    with pytest.raises(ValueError) as exc:
        defn = SchemaDefinition({"name": "test", "mapping": {}, "samples": ["42"]})
        defn.parse()
    assert str(exc.value) == "test: Invalid schema sample: 42"


def test_invalid_sample_length():
    with pytest.raises(ValueError) as exc:
        defn = SchemaDefinition({"name": "test", "mapping": {}, "samples": [["aa"]]})
        defn.parse()
    assert str(exc.value) == "test: Invalid schema sample: ['aa']"


def test_invalid_sample_type():
    with pytest.raises(ValueError) as exc:
        defn = SchemaDefinition({"name": "test", "mapping": {}, "samples": [["aa", {}]]})
        defn.parse()
    assert str(exc.value) == "test: Invalid schema sample: ['aa', {}]"


def test_missing_samples():
    defn = SchemaDefinition({"name": "test", "mapping": {}})
    defn.parse()
    assert defn.samples is None


def test_empty_samples():
    defn = SchemaDefinition({"name": "test", "mapping": {}, "samples": []})
    defn.parse()
    assert defn.samples == []



"""
Transliteration schema registry.
"""

import json
from operator import attrgetter
from pathlib import Path
from enum import Enum
from typing import List, Tuple
from iuliia.schema import Schema


def _schema_loader():
    return (Schema.load(defn) for defn in _definition_reader())


def _definition_reader():
    schemas_path = Path(__file__).parent / "schemas"
    if not schemas_path.exists():
        raise ValueError(f"Schema path does not exist: {schemas_path}")
    paths = schemas_path.glob("*.json")
    for path in paths:
        definition = _load_definition(path)
        yield definition


def _load_definition(path):
    with open(path, encoding="utf-8") as file:
        return json.load(file)


class _Schemas(Enum):
    """Base class for supported transliteration schemas."""

    @classmethod
    def names(cls) -> List[str]:
        """Return names of all supported schemas."""
        return sorted(item.name for item in cls)

    @classmethod
    def items(cls) -> List[Tuple[str, Schema]]:
        """Return all supported schemas."""
        return [(item.name, item.value) for item in sorted(cls, key=attrgetter("value.name"))]

    @classmethod
    def get(cls, name: str) -> Schema:
        """Return schema by its name or ``None`` if nothing found."""
        item = cls.__members__.get(name)
        return item.value if item else None


# All supported transliteration schemas
# pylint: disable=invalid-name
Schemas = Enum(  # type: ignore
    "Schemas", [(schema.name, schema) for schema in _schema_loader()], type=_Schemas
)



import iuliia


def test_init():
    schema = iuliia.Schema("test", {"a": "z"})
    assert schema.name == "test"
    assert schema.map.map == {"a": "z", "A": "Z"}


def test_str():
    schema = iuliia.Schema("test", {})
    assert str(schema) == "test"
    assert repr(schema) == "test"
